#include <sys/types.h>
#include <sys/ata.h>
#include <sys/ioctl.h>
#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define	STAT_GOOD_STR		"GOOD"
#define	STAT_PREFAILURE_STR	"PREFAILURE"
#define	STAT_BAD_STR		"BAD"

int main(int argc, char **argv)
{
	char indicator[51];
	char *status;
	int fd, i, j, overall, pos, rval, threshold;
	float ratio;
	struct ata_cmd cmd;
	struct ata_smart_vals vals;
	struct ata_smart_thrs thrs;
	FILE *file;
	char attrdesc[256][160];
	int attrcount;
	unsigned int attr;
	char desc[160];
	char line[160];
	char idstr[5];
	char *lptr, *strend;
	int len;

	attrcount = 0;
	for (i = 0; i < 256; i++)
	    bzero(attrdesc[i], 160);
	
	if ((file = fopen("SMART.txt", "r")) == NULL)
	    warn("Couldn't open SMART.txt for reading");
	else {
	    while (!feof(file)) {
		lptr = fgetln(file, &len);
		if (lptr == NULL)
		    continue;
		
		if (len > sizeof(line))
		    errx(4, "Line too long, %d > %d", len, sizeof(line));
		    
		if (lptr[0] == '#' || len == 1)
		    continue;

		if (lptr[len - 1] == '\n') {
		    len--;
		}
		
		memcpy(line, lptr, len);
		line[len] = '\0';
		
		if (sscanf(line, "%4s %159[^\n]", idstr, desc) != 2)
		    errx(4, "Unable to parse '%s'", line);
		
		attr = strtol(idstr, &strend, 16);
		
		if (strend == idstr)
		    err(4, "Unable to parse attribute ID '%s'", idstr);
		
		if (attr < 0 || attr > 255)
		    err(4, "Attrtibute value %d out of range", attr);
		
		strcpy(attrdesc[attr], desc);
	    }
	}
	    
	fd = open("/dev/ata", O_RDONLY);
	if (fd == -1)
		err(4, "/dev/ata");

	if (argc != 3) {
	    cmd.channel = 0;
	    cmd.device = 0;
	} else {
	    cmd.channel = atoi(argv[1]);
	    cmd.device = atoi(argv[2]);
	}

	cmd.cmd = ATAGSMARTVALS;
	rval = ioctl(fd, IOCATA, &cmd);
	if (rval == -1)
		err(4, "ioctl(ATAGSMARTVALS)");
	vals = cmd.u.smart_vals;

	cmd.cmd = ATAGSMARTTHRS;
	rval = ioctl(fd, IOCATA, &cmd);
	if (rval == -1)
		err(4, "ioctl(ATAGSMARTTHRS)");
	thrs = cmd.u.smart_thrs;

	overall = 0;
	printf("ID#  ATTRIBUTE THRESHOLD(+)/CURRENT(-)/WORST EVER(X)      + / - / X  CONDITION NAME\n");
	printf("---  --------------------------------------------------  ----------- --------- ----\n");
	for (i = 0; i < SMART_NATTRS; i++) {
		if (SMART_ATTR_UNUSED(vals.attributes[i]))
			continue;
		memset(indicator, '-', sizeof(indicator) - 1);
		indicator[sizeof(indicator) - 1] = '\0';
		threshold = -1;
		for (j = 0; j < SMART_NATTRS; i++) {
			if (vals.attributes[i].id == thrs.thresholds[i].id) {
				threshold = thrs.thresholds[i].value;
				break;
			}
		}
		if (threshold == -1) {
			/* Shouldn't happen */
			warnx("no threshold for the attribute #%d",
			      vals.attributes[i].id);
			abort();
		}
		ratio = (float)(sizeof(indicator) - 1)/
			vals.attributes[i].value;
		pos = ratio * threshold + 1;
		if (pos > (sizeof(indicator) - 1))
			pos = sizeof(indicator) - 1;
		memset(indicator, '+', pos);
		pos = ratio * vals.attributes[i].worst_value - 1;
		indicator[pos] = 'X';
		if (vals.attributes[i].value > threshold ||
		    threshold == SMART_THRS_ALWAYS_PASS) {
			status = STAT_GOOD_STR;
		} else if (vals.attributes[i].value == threshold) {
			status = STAT_PREFAILURE_STR;
			if (overall > -1) 
				overall = -1;
		} else {
			status = STAT_BAD_STR;
			if (overall > -2)
				overall = -2;
		}
		printf("%3d |%s| %3d/%3d/%3d    %s", vals.attributes[i].id,
		       indicator, threshold, vals.attributes[i].value,
		       vals.attributes[i].worst_value, status);
		if (attrdesc[vals.attributes[i].id][0] != '\0')
		    printf("   %s\n", attrdesc[vals.attributes[i].id]);
		else
		    printf("\n");
	}
	switch (overall) {
	case 0:
		status = STAT_GOOD_STR;
		break;

	case -1:
		status = STAT_PREFAILURE_STR;
		break;

	case -2:
		status = STAT_BAD_STR;
		break;
	}
	printf("\nOverall drive condition: %s\n", status);
	exit(0);
}
